﻿using System.Windows.Controls;
using System.Windows.Documents;

namespace Sketchables
{
  public abstract class TextParserBase
  {
    /// <summary>
    /// Inits the parser with the supported tags.
    /// </summary>
    /// <param name="tags"></param>
    protected TextParserBase(params char[] tags)
    {
      Tags = tags;
    }

    /// <summary>
    /// The tags that are going to be parsed.
    /// </summary>
    public char[] Tags { get; private set; }

    /// <summary>
    /// Parses a given string into segments.
    /// </summary>
    /// <param name="markup">The markup to be processed.</param>
    public TextBlock Parse(string markup)
    {
      TextBlock tb = new TextBlock();
      ParseSegments(ParseInternal(markup), markup, tb);
      return tb;
    }

    /// <summary>
    /// Processes a string of markup that may contain formatting directives,
    /// and returns a corresponding text segment.
    /// </summary>
    /// <param name="markup"></param>
    /// <returns></returns>
    protected virtual TextSegment ParseInternal(string markup)
    {
      var root = new TextSegment(null, 0, default(char));

      TextSegment currentSegment = root;

      int position = 0;
      while ((position = markup.IndexOfAny(Tags, position)) != -1)
      {
        if (position > 0 && markup[position - 1] == '\\')
        {
          //if the tag was escaped, just ignore it
          position++;
          continue;
        }

        char tag = markup[position];

        if (tag == currentSegment.SegmentTag)
        {
          //set the end index (includes the tag character)
          currentSegment.EndIndex = position;

          //close the current segment and move up
          currentSegment = currentSegment.Parent;
        }
        else
        {
          //open a child segment (start index includes the tag character)
          var child = new TextSegment(currentSegment, position, tag);
          currentSegment.ChildSegments.Add(child);

          //make the child the new segment
          currentSegment = child;
        }

        position++;
      }

      root.EndIndex = markup.Length - 1;

      //in case tags weren't closed (either by mistake or because they just aren't tags,
      //the end index isn't set and everything that comes later is regarded a child segment
      //of this (invalid) segment. Remove those entries and insert their childs into the parent
      RemoveInvalidEntries(root);

      return root;
    }

    /// <summary>
    /// Removes entries without closing tags, and assigns their entries
    /// to the parent segment. This basically corrects segments that aren't
    /// any (e.g. due to a single tag character in a regular string).
    /// </summary>
    /// <param name="current">The currently processed segment (invoked
    /// recursively).</param>
    protected virtual void RemoveInvalidEntries(TextSegment current)
    {
      for (int index = 0; index < current.ChildSegments.Count; index++)
      {
        TextSegment segment = current.ChildSegments[index];

        //recurse first
        RemoveInvalidEntries(segment);

        if (segment.EndIndex == -1)
        {
          current.ChildSegments.RemoveAt(index);
          current.ChildSegments.InsertRange(index, segment.ChildSegments);

          //readjust parent reference
          foreach (var cs in segment.ChildSegments)
          {
            cs.Parent = current;
          }
        }
      }
    }

    /// <summary>
    /// Parses the segments into text blocks.
    /// </summary>
    /// <param name="segment"></param>
    /// <param name="markup"></param>
    /// <returns>The root TextBlock that contains the actual formatting elements
    /// and text as inlines.</returns>
    internal void ParseSegments(TextSegment segment, string markup, TextBlock container)
    {
      int position = segment.ContentStartIndex;

      foreach (TextSegment child in segment.ChildSegments)
      {
        //handle the string between the current position and this child
        int delta = child.StartIndex - position;
        if (delta > 0)
        {
          //create text run
          string content = Unescape(markup.Substring(position, delta));
          var run = new Run {Text = content};
          FormatTextBlock(run, segment);
          container.Inlines.Add(run);
        }

        //pass control to the child in order to recursively handle its range
        ParseSegments(child, markup, container);

        //reposition after the segment
        position = child.EndIndex + 1;
      }

      //handle text between the last child and the remaining text of the segment. If we didn't
      //have any childs, our position is still the segment's start index
      if (position <= segment.ContentEndIndex)
      {
        int delta = segment.ContentEndIndex - position + 1;
        string content = Unescape(markup.Substring(position, delta));
        var run = new Run { Text = content };
        FormatTextBlock(run, segment);
        container.Inlines.Add(run);
      }
    }

    protected abstract void FormatTextBlock(Run run, TextSegment segment);

    /// <summary>
    /// Unescapes a given text.
    /// </summary>
    protected abstract string Unescape(string text);
  }
}