﻿using System;
using System.Collections.Generic;
using System.Windows.Documents;


namespace Sketchables
{
  /// <summary>
  /// Represents a segment that spans a given formatting within
  /// the content. A segment can have nested child segment, e.g.
  /// a chunk of italic text in a chunk of bold text.
  /// </summary>
  public class TextSegment
  {
    /// <summary>
    /// The child (sub) segments that span some additional formatting.
    /// </summary>
    public List<TextSegment> ChildSegments { get; private set; }

    /// <summary>
    /// The parent segment, if any. The root segment returns null.
    /// </summary>
    public TextSegment Parent { get; set; }

    /// <summary>
    /// The start index within the processed markup, including the
    /// formatting tag.
    /// </summary>
    public int StartIndex { get; set; }

    /// <summary>
    /// The end index within the processed markup, including the
    /// formatting tag. Defaults to -1, which indicates that the segment is
    /// not closed and thus removed again during processing.
    /// </summary>
    public int EndIndex { get; set; }

    /// <summary>
    /// The start index of the actual content. This is the <see cref="StartIndex"/>
    /// plus the length of the <see cref="SegmentTag"/>.
    /// </summary>
    public int ContentStartIndex
    {
      get { return IsRoot ? StartIndex : StartIndex + 1; }
    }

    /// <summary>
    /// The start index of the actual content. This is the <see cref="EndIndex"/>
    /// minus the length of the <see cref="SegmentTag"/>.
    /// </summary>
    public int ContentEndIndex
    {
      get { return IsRoot ? EndIndex : EndIndex - 1; }
    }

    /// <summary>
    /// The tag that defines the segment (and formatting).
    /// </summary>
    public char SegmentTag { get; set; }

    /// <summary>
    /// Whether the segment is the root.
    /// </summary>
    public bool IsRoot
    {
      get { return Parent == null; }
    }

    /// <summary>
    /// Intializes a new segment instace with defaults.
    /// </summary>
    /// <param name="parent">The segment's parent, if it's a nested segment.</param>
    /// <param name="startIndex">The start index of the segment within the processed
    /// text.</param>
    /// <param name="segmentTag">The segment tag.</param>
    public TextSegment(TextSegment parent, int startIndex, char segmentTag)
    {
      Parent = parent;
      StartIndex = startIndex;
      EndIndex = -1;
      SegmentTag = segmentTag;

      ChildSegments = new List<TextSegment>();
    }


    /// <summary>
    /// Gets a hierarchical list of all text segments starting from this
    /// current instance up to the top.
    /// </summary>
    public IEnumerable<TextSegment> GetParentGraph()
    {
      TextSegment current = this;
      while (current != null)
      {
        yield return current;
        current = current.Parent;
      }
    }
  }
}
